import os
import sys
import importlib
import numpy as np
import torch
import time
import matplotlib.patches as patches
import cv2 as cv
import matplotlib.pyplot as plt
from pytracking.analysis.plot_results import get_plot_draw_styles
from pytracking.utils.plotting import draw_figure
from pytracking.evaluation import get_dataset, trackerlist, Tracker

env_path = os.path.join(os.path.dirname(__file__), '../..')
if env_path not in sys.path:
    sys.path.append(env_path)


class Display:
    def __init__(self, sequence_length, plot_draw_styles, sequence_name):
        self.active = True
        self.frame_number = 0
        self.pause_mode = False
        self.step_size = 0
        self.step_direction = 'forward'
        self.fig, self.ax = plt.subplots(1)
        self.fig.canvas.mpl_connect('key_press_event', self.key_callback_fn)
        plt.tight_layout()

        self.sequence_length = sequence_length
        self.sequence_name = sequence_name
        self.plot_draw_styles = plot_draw_styles

    def key_callback_fn(self, event):
        if event.key == ' ':
            self.pause_mode = not self.pause_mode
            self.step_size = 0
            self.step_direction = 'forward'
        elif event.key == 'right':
            if self.pause_mode:
                self.frame_number += 1

                if self.frame_number >= self.sequence_length:
                    self.frame_number = self.sequence_length - 1
            elif self.step_direction == 'stop':
                self.step_direction = 'forward'
                self.step_size = 0
            elif self.step_direction == 'backward' and self.step_size == 0:
                self.step_direction = 'stop'
            else:
                self.step_size += 1
        elif event.key == 'left':
            if self.pause_mode:
                self.frame_number -= 1

                if self.frame_number < 0:
                    self.frame_number = 0
            elif self.step_direction == 'stop':
                self.step_direction = 'backward'
                self.step_size = 0
            elif self.step_direction == 'forward' and self.step_size == 0:
                self.step_direction = 'stop'
            else:
                self.step_size -= 1
        elif event.key == 'escape' or event.key == 'q':
            self.active = False

    def _get_speed(self):
        delta = 0
        if self.step_direction == 'forward':
            delta = 2 ** abs(self.step_size)
        elif self.step_direction == 'backward':
            delta = -1 * 2 ** abs(self.step_size)

        return delta

    def step(self):
        delta = self._get_speed()

        self.frame_number += delta
        if self.frame_number < 0:
            self.frame_number = 0
        elif self.frame_number >= self.sequence_length:
            self.frame_number = self.sequence_length - 1

    def show(self, image, bb, tracker_name, gt=None, mask=None):
        self.ax.cla()

        if mask is not None:
            alpha = 0.5
            foreground = image * alpha + (1 - alpha) * mask
            image[(mask > 0).any(axis=2), :] = foreground[(mask > 0).any(axis=2), :]

        self.ax.imshow(image)

        # Draw rects
        rect_handles = []
        rect = patches.Rectangle((bb[0], bb[1]), bb[2], bb[3], linewidth=1,
                                 edgecolor=self.plot_draw_styles[0]['color'], facecolor='none')
        self.ax.add_patch(rect)

        rect_handles.append(patches.Rectangle((bb[0], bb[1]), bb[2], bb[3], linewidth=1,
                                              edgecolor=self.plot_draw_styles[0]['color'],
                                              facecolor=self.plot_draw_styles[0]['color'],
                                              label=tracker_name))

        if gt is not None:
            rect = patches.Rectangle((gt[0], gt[1]), gt[2], gt[3], linewidth=2, edgecolor='g',
                                     facecolor='none')
            self.ax.add_patch(rect)
            rect_handles.append(rect)

        self.ax.set_axis_off()
        self.ax.axis('equal')
        plt.legend(handles=rect_handles, loc=4, borderaxespad=0.)
        mode = 'manual' if self.pause_mode else 'auto     '
        speed = self._get_speed()
        self.fig.suptitle('Sequence: {}    Mode: {}    Speed: {:d}x'.format(self.sequence_name, mode, speed),
                          fontsize=14)
        draw_figure(self.fig)


def read_image(image_file: str):
    im = cv.imread(image_file)
    return cv.cvtColor(im, cv.COLOR_BGR2RGB)


def _get_display_name(tracker):
    if tracker.display_name is None:
        if tracker.run_id is not None:
            return '{}_{}_{:03d}'.format(tracker.name, tracker.parameter_name, tracker.run_id)
        else:
            return '{}_{}'.format(tracker.name, tracker.parameter_name)
    else:
        return tracker.display_name


def playback_results(tracker, sequence):

    plot_draw_styles = get_plot_draw_styles()

    # Load results
    base_results_path = '{}/{}'.format(tracker.results_dir, sequence.name)
    results_path = '{}.txt'.format(base_results_path)

    if os.path.isfile(results_path):
        try:
            pred_bb = torch.tensor(np.loadtxt(str(results_path), dtype=np.float64))
        except:
            pred_bb = torch.tensor(np.loadtxt(str(results_path), delimiter=',', dtype=np.float64))
    else:
        raise Exception('Result not found. {}'.format(results_path))

    # Load masks if available
    masks_path = '{}/{}'.format(tracker.segmentation_dir, sequence.name)
    if os.path.isdir(masks_path):
        segmentation_masks = [cv.imread('{}/{}.png'.format(masks_path, os.path.basename(f)[:-4]))
                              for f in sequence.frames]
    else:
        segmentation_masks = [None]*pred_bb.shape[0]

    # Convert to list of shape seq_length * num_trackers * 4
    tracker_results = pred_bb.tolist()
    tracker_name = _get_display_name(tracker)

    display = Display(len(tracker_results), plot_draw_styles, sequence.name)

    while display.active:
        frame_number = display.frame_number
        image = read_image(sequence.frames[frame_number])

        display.show(image, tracker_results[frame_number], tracker_name, mask=segmentation_masks[frame_number])

        time.sleep(0.01)
        if display.pause_mode and display.frame_number == frame_number:
            time.sleep(0.1)
        elif not display.pause_mode:
            display.step()


if __name__ == '__main__':
    tracker = Tracker('dimp_vos', 'res50_vos_mrcnn_lwt_got10k_lasot_ft_ep120', 0, 'LWL-Ours')
    dataset = get_dataset('got10k_val')

    playback_results(tracker, dataset[13])
